/**
 */
package LPCforSOS.evaluation;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instances;
import LPCforSOS.baseLearner.DecomposedDatasetLearner;
import LPCforSOS.dataStructure.ExtInstances;
import LPCforSOS.dataStructure.ExtInstancesOnARFF;
import LPCforSOS.decomposition.Decomposer;
import LPCforSOS.structuredPrediction.AggregationType;
import LPCforSOS.structuredPrediction.StructuredPredictor;
import LPCforSOS.structuredPrediction.WeightedAndUnweightedVoting;

/**
 * Evaluierung auf den Originaldaten (earff): Daten werden zunaechst in n folds
 * aufgeteilt und danach erst in paarweise Lernprobleme aufgespaltet.
 * 
 * @author Sandra Ebert, George-P. C.-F.
 * 
 * @date 08.04.2008, 08:22:15
 */
public class MovieEvaluation {

	private static String outputFilePath;
	private static String inputFilePath;
	Result results;

	/**
	 * @param results
	 */
	public MovieEvaluation(Result results) {
		this.results = results;
	}

	/**
	 * Crossvalidation
	 * 
	 * @throws Exception
	 */
	public void crossValidateModel() throws Exception {

		

		results.initialize();

		// Do the folds
		for (int fold = 0; fold < results.getFolds(); fold++) {
			ExtInstances data = ((FixedCVResult)results).getTrainInstances().get(0);
			
			long time = System.currentTimeMillis();

			// decompose train set
			Decomposer decomposition = Decomposer.forName(results
					.getDecompositionType());
			decomposition.setSymmetric(results.isSymmetric());
			ExtInstances trainingSet = ((FixedCVResult)results).getTrainInstances().get(fold);
			decomposition.buildDecomposedDatasets(trainingSet);

			// build the base classifiers for decomposed datasets
			DecomposedDatasetLearner models = DecomposedDatasetLearner
					.forName(results.getEnvironment());
			models.setDecomposedDatasets(decomposition.getDecomposedDatasets());
			models.buildClassifiers(results.getClassifierName()); // spark:
																	// thats
																	// strange,
																	// remodel
																	// that
			models.setClassToNumber(decomposition.getClassToNumber());

			// decompose test set
			ExtInstances extInstances = ((FixedCVResult)results).getTestInstances().get(fold);
			extInstances.setTestSetClasses(new ArrayList<List<String>>());
			extInstances.setTestSetTotalOrderOfClasses(new ArrayList<List<String>>());
			FastVector attributes = (FastVector) extInstances.get_Attributes()
					.copyElements();
			FastVector labels = new FastVector();
			Set<String> testClasses  = decomposition.getClassToNumber().keySet();
			// use classes from file
			if (testClasses.size() == 0) {
				for (String item : extInstances.getClasses()) {
					labels.addElement(item);
				}
			} else {
				for (String item : testClasses) {
					labels.addElement(item);
				}
			}
			attributes.addElement(new Attribute("class", labels));

			String sub_relation = extInstances.relationName() + "_test";
			
			
			Instances test = new Instances(sub_relation,attributes,extInstances.numInstances());
			for (int i = 0; i < extInstances.numInstances(); i++) {
				test.add(extInstances.instance(i, test));
			}
			test.setClassIndex(test.numAttributes()-1);
			// Structured Prediction with base classifiers
			StructuredPredictor pred = StructuredPredictor.forName(results);
			pred.classifyInstances(test, models, trainingSet);

			// record the time
			results.addFoldTime(System.currentTimeMillis() - time);

			/********** Datenerfassung fuer die Auswertung */

			String[] allLabelsArray = new String[models.getClassToNumber()
					.size()];
			// Hier werden die Zeichenketten gespeichert fuer die gevoted
			// wurden. Sind z.B beim labelRanking Relationen (a<b).
			ArrayList<String> votedStrings = new ArrayList<String>(models
					.getClassToNumber().size());
			for (String label : models.getClassToNumber().keySet()) {
				allLabelsArray[models.getClassToNumber().get(label)] = label;
			}
			for (int i = 0; i < allLabelsArray.length; i++) {
				votedStrings.add(allLabelsArray[i]);
			}

			// Hier werden alle existierenden Labels gespeichert.
			List<String> allLabels = data.getClasses();

			// Enthaelt alle votes die jede der Zeichenketten pro Instanz
			// bekommen hat.
			ArrayList<double[]> listOfVotesPerInstance = pred.getVoteVector();

			ArrayList<double[]> listOfDistributionsPerInstance = null;
			if(Configuration.aggregationType == AggregationType.WeightedAndUnweightedVoting){
				listOfVotesPerInstance = ((WeightedAndUnweightedVoting) pred).getDistributionVector();
			}
			
			// Enthaelt alle relevanten Labels pro Instanz.
			List<List<String>> relevantLabels = extInstances.getTestSetClasses();

			// Bei LabelRanking ist hier die Totale Ordung der Labels
			// gespeichert.
			List<List<String>> labelRankingTotalOrder = data
					.getTestSetTotalOrderOfClasses();

			/********** Ende Datenerfassung */

			results.addFoldResult("Fold " + (fold + 1), allLabels,
					votedStrings, relevantLabels, listOfVotesPerInstance,
					labelRankingTotalOrder,listOfDistributionsPerInstance);
			// results.getFoldResults().get("Fold 0").evaluate();
		}

		// Ausfuehrung der Evaluation f�r die einzelnen Fold Ergebnisse
		for (String s : results.getFoldResults().keySet()) {
			results.getFoldResults().get(s).evaluate();
		}

		// Ausgabe der Ergebnisse
		System.out.println(results);
		Writer w = new BufferedWriter(new FileWriter(outputFilePath, true));
		w.write(results.toString());
		w.flush();
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String mode = args[0];
		if("t".equals(mode)){
			Configuration.aggregationType = AggregationType.WeightedAndUnweightedVoting;
		} else if("w".equals(mode)) {
			Configuration.aggregationType = AggregationType.WeightedVoting;
		} else {
			throw new IllegalArgumentException("use 'w' for weighted voting and 't' for tie breaking with weighted voting");
		}
		
		inputFilePath = args[1];
		outputFilePath = args[2];
		try {
			List<ExtInstances> foldsTrainData = new ArrayList<ExtInstances>();
			extractFolds(foldsTrainData, true);
			List<ExtInstances> foldsTestData = new ArrayList<ExtInstances>();
			extractFolds(foldsTestData, false);

			// ******** TODO Nutzlos den zu setzen da in earff Datei geschrieben
			// Classification Attribute Type
			// Type type = Type.LabelRanking;

			// Result Objekt fuer die entsprechende Klassifikation
			Result results = new FixedCVResult(foldsTrainData, foldsTestData); // Result.forName(type);
			// results.setClassificationAttributeType(type);

			// Von welchem Typ ist das Klassen Attribut.
			// results.setClassAttributeType(data.classAttribute().type());

			// symmetric classifier; false if asymmetric -> double round robin
			results.setSymmetric(Configuration.isSymmetricClassifier);

			// Decomposition
			results.setDecompositionType(Configuration.decompositionType);

			// Aggregation
			results.setAggregationType(Configuration.aggregationType);

			// Classifier Environment
			results.setBaseLearnerEnvironment(Configuration.baseLearnerEnvironment);

			// Classifier
			results.setClassifier(Configuration.positionOfClassifier);

			// // Crossvalidate the base classifiers
			// results.setEvaluateBaseClassifiers(Configuration.evaluateBaseClassifiers);

			// Seed for evaluation
			results.setRandom(Configuration.seedForEvaluation);

			// Number of folds
			results.setFolds(Configuration.numberOfFolds);

			// Crossvalidation method
			MovieEvaluation eval = new MovieEvaluation(results);
			eval.crossValidateModel();

		} catch (Exception ex) {
			ex.printStackTrace();
			System.err.println(ex.getMessage());
		}
	}

	private static void extractFolds(List<ExtInstances> foldsTrainData,
			boolean isTrainingSet) throws FileNotFoundException, IOException {
		String testSetKey = "test";
		String trainingSetKey = "train";
		String setKey = (isTrainingSet) ? trainingSetKey : testSetKey;
		Reader r;
		for (int i = 0; i < Configuration.numberOfFolds; i++) {
			Configuration.positionOfEARFFFile = inputFilePath + "movielist_CV"
					+ (i + 1) + "-10_" + setKey + ".earff";
			System.out.println(setKey + " " + (i + 1) + "\n");
			r = new BufferedReader(new FileReader(
					Configuration.positionOfEARFFFile));

			ExtInstances data = null;
			// Abfrage ob weka arff Datei.
			if (Configuration.positionOfEARFFFile.contains(".arff")) {
				data = new ExtInstancesOnARFF(r);
			} else {
				data = new ExtInstances(r);
			}

			foldsTrainData.add(data);
		}
	}

	/**
	 * Validierungsset wird zufaellig ausgewaehlt. (Weka Cross Validation)
	 * 
	 * @param numfolds
	 *            gibt Groesse des Validierungssets an
	 */
	public void getHoldoutValidationSet(int numfolds) {

	}
}
